@@ -427,24 +427,6 @@ struct terminal_tag {
         WIN_RESIZE_NO, WIN_RESIZE_NEED_SEND, WIN_RESIZE_AWAIT_REPLY
     } win_resize_pending;
     int win_resize_pending_w, win_resize_pending_h;
-
-    /*
-     * Not every frontend / TermWin implementation can be relied on
-     * 100% to reply to a resize request in a timely manner. (In X11
-     * it's all asynchronous and goes via the window manager, and if
-     * your window manager is seriously unwell, you'd rather not have
-     * terminal windows start becoming unusable as a knock-on effect,
-     * since those are just the thing you might need to use for
-     * emergency WM maintenance!) So when we enter AWAIT_REPLY status,
-     * we also set a 5-second timer, after which we'll regretfully
-     * conclude that a resize is probably not going to happen after
-     * all.
-     *
-     * However, in non-emergency cases, the plan is that this
-     * shouldn't be needed, for one reason or another.
-     */
-    long win_resize_timeout;
-    #define WIN_RESIZE_TIMEOUT (TICKSPERSEC*5)
 };
 
 static inline bool in_utf(Terminal *term)
