@@ -12,8 +12,7 @@ void align_label_left(GtkLabel *label);
 
 GtkWidget *our_dialog_new(void);
 void our_dialog_add_to_content_area(GtkWindow *dlg, GtkWidget *w,
-                                    gboolean expand, gboolean fill,
-                                    guint padding);
+                                    bool expand, bool fill, guint padding);
 void our_dialog_set_action_area(GtkWindow *dlg, GtkWidget *w);
 GtkBox *our_dialog_make_action_hbox(GtkWindow *dlg);
 
